@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1996-2014 The Squid Software Foundation and contributors
+ * Copyright (C) 1996-2017 The Squid Software Foundation and contributors
  *
  * Squid software is distributed under GPLv2+ license and includes
  * contributions from numerous individuals and organizations.
@@ -48,11 +48,13 @@ public:
 
 /*******************************************************************
  * external_acl cache entry
- * Used opaqueue in the interface
+ * Used opaque in the interface
  */
 
 class ExternalACLEntry: public hash_link, public RefCountable
 {
+    MEMPROXY_CLASS(ExternalACLEntry);
+
 public:
     ExternalACLEntry();
     ~ExternalACLEntry();
@@ -73,11 +75,7 @@ public:
     String tag;
     String log;
     external_acl *def;
-
-    MEMPROXY_CLASS(ExternalACLEntry);
 };
 
-MEMPROXY_CLASS_INLINE(ExternalACLEntry);
-
 #endif
 
