@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1996-2014 The Squid Software Foundation and contributors
+ * Copyright (C) 1996-2017 The Squid Software Foundation and contributors
  *
  * Squid software is distributed under GPLv2+ license and includes
  * contributions from numerous individuals and organizations.
@@ -12,7 +12,7 @@
 #include "acl/Checklist.h"
 #include "acl/Data.h"
 #include "acl/Strategised.h"
-#include "CbDataList.h"
+#include "base/CbDataList.h"
 #include "ip/Address.h"
 
 int asnMatchIp(CbDataList<int> *, Ip::Address &);
@@ -26,10 +26,10 @@ void asnFreeMemory(void);
 /// \ingroup ACLAPI
 class ACLASN : public ACLData<Ip::Address>
 {
-
-public:
     MEMPROXY_CLASS(ACLASN);
 
+public:
+    ACLASN() : data(nullptr) {}
     virtual ~ACLASN();
 
     virtual bool match(Ip::Address);
@@ -47,7 +47,5 @@ private:
     CbDataList<int> *data;
 };
 
-MEMPROXY_CLASS_INLINE(ACLASN);
-
 #endif /* SQUID_ACLASN_H */
 
